/*
 * $Id$
 * Copyright (C) 2001 The Apache Software Foundation. All rights reserved.
 * For details on use and redistribution please refer to the
 * LICENSE file included with these sources.
 */

package org.apache.fop.fonts;

import org.apache.fop.apps.FOPException;
import org.apache.fop.layout.FontMetric;
import org.apache.fop.layout.FontDescriptor;
import org.apache.fop.render.pdf.CodePointMapping;
import org.apache.fop.messaging.MessageHandler;

import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * Encapsulates information for typefaces of interest to FOP,
 * regardless of their source.
 * Within FOP, a Font is a specific instance of the more general Typeface,
 * which is itself a member of a TypefaceFamily.
 * A Font is a Typeface rendered at a specific size, and perhaps with
 * various transformations applied.
 * Please note that the terms "font" and "typeface" are used interchangeably in
 * many other contexts. Scaleable font files, for example, usually contain
 * typeface data, or data for an entire family of typefaces.
 *
 * Typeface data is currently derived from two sources:
 * <ul>
 *   <li>
 *     Hard-Coded. These are the base-14 fonts. Since they never need
 *     to be embedded, there is no need for access to the physical font file.
 *     However, the metrics are hard-coded within FOP.
 *   <li>
 *     Direct Access (DA). These typefaces are accessed by FOP through its
 *     configuration, which points to an XML file containing the font
 *     metrics, and (where needed) the physical font file.
 *     These font metrics must be parsed by FOP routines before they can be
 *     used by FOP.
 *     DA Fonts may or may not be registered at the operating
 *     system level, but if they are, such registration is ignored.
 *   </li>
 * </ul>
 *
 * Planned enhancements to this class include addition of support for
 * Indirect Access (IA) Fonts. These fonts are always registered at the
 * operating system level, and are available to FOP through AWT classes.
 * Please note that references to IA Fonts may still appear
 * in the FOP configuration, but only for purposes such as embedding them
 * in rendered output.
 *
 * Support for Direct Access Fonts is desirable because of the following:
 * <ul>
 *   <li>
 *     There is at least some doubt about whether we can fully trust AWT for
 *     layout chores.
 *   </li>
 *   <li>
 *     There is a desire for backward compatibility.
 *   </li>
 * </ul>
 *
 * Support for Indirect Access Fonts is desirable because of the following:
 * <ul>
 *   <li>
 *     It immediately gives FOP access to all font types (including OpenType)
 *     that java supports.
 *   </li>
 *   <li>
 *     It gives us a way to test AWT without breaking preexisting functionality.
 *   </li>
 *   <li>
 *     It is a generally cleaner way to handle fonts, both for users and
 *     developers.
 *   </li>
 * </ul>
 *
 * @author Unascribed
 * @author {@link <a href="mailto:vic@outfitr.com">Victor Mote</a>}
 * refactored existing code into this class.
 */
public class Typeface {

    /**
     * HashMap of typefaces used.
     * Not sure whether the content is different from tripletsFontsMap and
     * fontsMap, or whether only the keys are different, allowing multiple ways to
     * find the same font.
     */
    public static HashMap usedFontsMap = new HashMap();

    /**
     * HashMap of fonts keyed by the font-triplet (family, style, and weight).
     * Not sure whether the content is different from tripletsFontsMap and
     * fontsMap, or whether only the keys are different, allowing multiple ways to
     * find the same font.
     */
    public static HashMap tripletsFontsMap = new HashMap();

    /**
     * HashMap of fonts keyed by font name. May only include fonts that implement
     * FontMetric.
     * Not sure whether the content is different from tripletsFontsMap and
     * fontsMap, or whether only the keys are different, allowing multiple ways to
     * find the same font.
     */
    public static HashMap fontsMap = new HashMap();

    /** The font family to which this font belongs. */
    private String fontFamily;

    /**
     * The FO typeface style (italic, etc.). Contents should conform to the
     * XSL-FO standard.
     */
    private String foStyle;

    /**
     * The FO typeface weight (bold, semibold, etc.). Contents should
     * conform to the XSL-FO standard.
     */
    private String foWeight;

    /**
     * The FO typeface variant (normal, small-caps).
     */
    private int foVariant;

    /** The Font Metric object for this typeface. */
    private FontMetric fontMetric;

    /** An empty hashmap used for kerning information */
    private static HashMap EMPTY_HASHMAP = new HashMap();

    /** @return the name of the Font Family for this typeface. */
    public String getFamily(){
        return fontFamily;
    }

    /** @return the font style for this typeface. */
    public String getFOStyle(){
        return foStyle;
    }

    /** Returns the font weight for this typeface. */
    public String getFOWeight(){
        return foWeight;
    }

    /** Returns the font variant for this typeface. */
    public int getFOVariant(){
        return foVariant;
    }

    /** Returns the fontMetric object for this typeface. */
    public FontMetric getFontMetric() {
        return fontMetric;
    }

    public HashMap getKerning() {
        if (fontMetric instanceof FontDescriptor) {
            HashMap ret = ((FontDescriptor)fontMetric).getKerningInfo();
            if (ret != null)
                return ret;
        }
        return EMPTY_HASHMAP;
    }

    /**
     * Maps a java character (unicode) to a font character.
     * Default uses CodePointMapping.
     */
    public char mapChar(char c) {
        if (fontMetric instanceof org.apache.fop.render.pdf.Font) {
            return ((org.apache.fop.render.pdf.Font)fontMetric).mapChar(c);
        } else if (fontMetric instanceof org.apache.fop.render.awt.FontMetricsMapper) {
            return c;
        }
        // Use default CodePointMapping
        char d = CodePointMapping.getMapping("WinAnsiEncoding").mapChar(c);
        if (d != 0) {
            c = d;
        } else {
            c = '#';
	    }
        return c;
    }

/*////////////////// METHODS FOR MANAGING THE FONT MAPS /////////////////*/

    /** Adds an entry to the tripletsFontsMap (uses family, style and weight as key. */
    public static void addFontProperties(String name, String family,
            String style, String weight) {
        String key = createFontKey(family, style, weight);
        tripletsFontsMap.put(key, name);
    }

    /** Creates a key for the tripletsFontsMap from the given strings */
    public static String createFontKey(String family, String style,
                                       String weight) {
        int i;
        try {
            i = Integer.parseInt(weight);
        } catch (NumberFormatException e) {
            i = 0;
        }
        if (i > 600)
            weight = "bold";
        else if (i > 0)
            weight = "normal";
        return family + "," + style + "," + weight;
    }

    /**
     * Looks for alternative font family names that might be in the
     * tripletsFontsMap.
     *
     * @return name of font family found in table, or null if not found.
     */
    public static String findFontFamilyNameToUse
            (String family, String style, String weight) {
        String fontKey = createFontKey(family, style, weight);
        if (hasFont(fontKey)) {
            return family;
        }
        //Tokenizes font-family list
        StringTokenizer st = new StringTokenizer(family, ",");
        while (st.hasMoreTokens()) {
            String token = st.nextToken().trim();
            token = TypefaceFamily.normalizeTypefaceFamilyName(token);
            //Checks found font family name for existence
            fontKey = createFontKey(token, style, weight);
            if (hasFont(fontKey)) {
                return family;
            }
        }
        return null;
    }

    /** Add the given metrics as a font with the given name. */
    public static void addMetrics(String name, FontMetric metrics) {
        fontsMap.put(name, metrics);
    }

    public static String fontLookup(String family, String style,
                             String weight) throws FOPException {
        return fontLookup(createFontKey(family, style, weight));
    }

    public static String fontLookup(String key) throws FOPException {

        String f = (String)tripletsFontsMap.get(key);
        if (f == null) {
            int i = key.indexOf(',');
            String s = "any" + key.substring(i);
            f = (String)tripletsFontsMap.get(s);
            if (f == null) {
                f = (String)tripletsFontsMap.get("any,normal,normal");
                if (f == null) {
                    throw new FOPException("no default font defined by OutputConverter");
                }
                MessageHandler.errorln("defaulted font to any,normal,normal");
            }
            MessageHandler.errorln("unknown font " + key
                                   + " so defaulted font to any");
        }

        usedFontsMap.put(f, fontsMap.get(f));
        return f;
    }

    public static boolean hasFont(String family, String style, String weight) {
        String key = createFontKey(family, style, weight);
        return tripletsFontsMap.get(key) != null;
    }

    public static boolean hasFont(String key) {
        return tripletsFontsMap.get(key) != null;
    }

    public static HashMap getFonts() {
        return fontsMap;
    }

    public static HashMap getUsedFonts() {
        return usedFontsMap;
    }

    public static FontMetric getMetricsFor(String fontName) throws FOPException {
        getUsedFonts().put(fontName, fontsMap.get(fontName));
        return (FontMetric)fontsMap.get(fontName);
    }

}

/* Last Line of $RCSfile$ */
